@@ -1,20 +1,21 @@
+use crate::Engine;
 use crate::{
     signatures::SignatureCollection,
     types::{ExportType, ExternType, ImportType},
 };
-use crate::{Engine, ModuleType};
 use anyhow::{bail, Context, Result};
 use once_cell::sync::OnceCell;
 use std::fs;
 use std::mem;
+use std::ops::Range;
 use std::path::Path;
 use std::sync::Arc;
 use wasmparser::{Parser, ValidPayload, Validator};
 use wasmtime_environ::{
-    DefinedFuncIndex, DefinedMemoryIndex, FunctionInfo, ModuleEnvironment, ModuleIndex, PrimaryMap,
-    SignatureIndex,
+    DefinedFuncIndex, DefinedMemoryIndex, FunctionInfo, ModuleEnvironment, PrimaryMap,
+    SignatureIndex, TypeTables,
 };
-use wasmtime_jit::{CompiledModule, CompiledModuleInfo, TypeTables};
+use wasmtime_jit::{CompiledModule, CompiledModuleInfo};
 use wasmtime_runtime::{
     CompiledModuleId, MemoryImage, MmapVec, ModuleMemoryImages, VMSharedSignatureIndex,
 };
@@ -103,20 +104,17 @@ struct ModuleInner {
     /// The compiled artifacts for this module that will be instantiated and
     /// executed.
     module: Arc<CompiledModule>,
-    /// Closed-over compilation artifacts used to create submodules when this
-    /// module is instantiated.
-    artifact_upvars: Vec<Arc<CompiledModule>>,
-    /// Closed-over module values which are used when this module is
-    /// instantiated.
-    module_upvars: Vec<Module>,
-    /// Type information of this module and all `artifact_upvars` compiled
-    /// modules.
+    /// Type information of this module.
     types: Arc<TypeTables>,
     /// Registered shared signature for the module.
     signatures: Arc<SignatureCollection>,
-    /// A set of initialization images for memories, if any. Note that module
-    /// instantiation (hence the need for lazy init) may happen for the same
-    /// module concurrently in multiple Stores, so we use a OnceCell.
+    /// A set of initialization images for memories, if any.
+    ///
+    /// Note that this is behind a `OnceCell` to lazily create this image. On
+    /// Linux where `memfd_create` may be used to create the backing memory
+    /// image this is a pretty expensive operation, so by deferring it this
+    /// improves memory usage for modules that are created but may not ever be
+    /// instantiated.
     memory_images: OnceCell<Option<ModuleMemoryImages>>,
 }
 
@@ -295,23 +293,14 @@ impl Module {
     #[cfg(compiler)]
     #[cfg_attr(nightlydoc, doc(cfg(feature = "cranelift")))] // see build.rs
     pub fn from_binary(engine: &Engine, binary: &[u8]) -> Result<Module> {
-        // Check to see that the config's target matches the host
-        let target = engine.compiler().triple();
-        if *target != target_lexicon::Triple::host() {
-            bail!(
-                "target '{}' specified in the configuration does not match the host",
-                target
-            );
-        }
-
-        // FIXME: we may want to validate that the ISA flags in the config match those that
-        // would be inferred for the host, otherwise the JIT might produce unrunnable code
-        // for the features the host's CPU actually has.
+        engine
+            .check_compatible_with_native_host()
+            .context("compilation settings are not compatible with the native host")?;
 
         cfg_if::cfg_if! {
             if #[cfg(feature = "cache")] {
                 let state = (HashedEngineCompileEnv(engine), binary);
-                let (main_module, artifacts, types) = wasmtime_cache::ModuleCacheEntry::new(
+                let (mmap, info, types) = wasmtime_cache::ModuleCacheEntry::new(
                     "wasmtime",
                     engine.cache_config(),
                 )
@@ -322,40 +311,28 @@ impl Module {
                     |(engine, wasm)| Module::build_artifacts(engine.0, wasm),
 
                     // Implementation of how to serialize artifacts
-                    |(engine, _wasm), (_, artifacts, types)| {
+                    |(engine, _wasm), (mmap, _info, types)| {
                         SerializedModule::from_artifacts(
                             engine.0,
-                            artifacts.iter().map(|p| &p.0),
+                            mmap,
                             types,
                         ).to_bytes(&engine.0.config().module_version).ok()
                     },
 
                     // Cache hit, deserialize the provided artifacts
                     |(engine, _wasm), serialized_bytes| {
-                        let (i, m, t, upvars) = SerializedModule::from_bytes(&serialized_bytes, &engine.0.config().module_version)
+                        SerializedModule::from_bytes(&serialized_bytes, &engine.0.config().module_version)
                             .ok()?
                             .into_parts(engine.0)
-                            .ok()?;
-                        // This upvars list is always empty for top-level modules
-                        assert!(upvars.is_empty());
-                        Some((i, m, t))
+                            .ok()
                     },
                 )?;
             } else {
-                let (main_module, artifacts, types) = Module::build_artifacts(engine, binary)?;
+                let (mmap, info, types) = Module::build_artifacts(engine, binary)?;
             }
         };
 
-        let modules = engine.run_maybe_parallel(artifacts, |(a, b)| {
-            CompiledModule::from_artifacts(
-                a,
-                b,
-                &*engine.config().profiler,
-                engine.unique_id_allocator(),
-            )
-        })?;
-
-        Self::from_parts(engine, modules, main_module, Arc::new(types), &[])
+        Self::from_parts(engine, mmap, info, Arc::new(types))
     }
 
     /// Converts an input binary-encoded WebAssembly module to compilation
@@ -379,81 +356,62 @@ impl Module {
     pub(crate) fn build_artifacts(
         engine: &Engine,
         wasm: &[u8],
-    ) -> Result<(
-        usize,
-        Vec<(MmapVec, Option<CompiledModuleInfo>)>,
-        TypeTables,
-    )> {
+    ) -> Result<(MmapVec, Option<CompiledModuleInfo>, TypeTables)> {
         let tunables = &engine.config().tunables;
 
         // First a `ModuleEnvironment` is created which records type information
         // about the wasm module. This is where the WebAssembly is parsed and
         // validated. Afterwards `types` will have all the type information for
         // this module.
-        let (main_module, translations, types) =
-            ModuleEnvironment::new(tunables, &engine.config().features)
-                .translate(wasm)
-                .context("failed to parse WebAssembly module")?;
+        let (mut translation, types) = ModuleEnvironment::new(tunables, &engine.config().features)
+            .translate(wasm)
+            .context("failed to parse WebAssembly module")?;
 
-        // Perform a two-level map/reduce here to get the final list of
-        // compilation artifacts. The first level of map/reduce maps over all
-        // modules found and reduces to collection into a vector. The second
-        // level of map/reduce here maps over all functions within each wasm
-        // module found and collects into an ELF image via `emit_obj`.
-        let list = engine.run_maybe_parallel(translations, |mut translation| -> Result<_> {
-            let functions = mem::take(&mut translation.function_body_inputs);
-            let functions = functions.into_iter().collect::<Vec<_>>();
-
-            let funcs = engine
-                .run_maybe_parallel(functions, |(index, func)| {
-                    engine
-                        .compiler()
-                        .compile_function(&translation, index, func, tunables, &types)
-                })?
-                .into_iter()
-                .collect();
-
-            let mut obj = engine.compiler().object()?;
-            let (funcs, trampolines) =
+        // Next compile all functions in parallel using rayon. This will perform
+        // the actual validation of all the function bodies.
+        let functions = mem::take(&mut translation.function_body_inputs);
+        let functions = functions.into_iter().collect::<Vec<_>>();
+        let funcs = engine
+            .run_maybe_parallel(functions, |(index, func)| {
                 engine
                     .compiler()
-                    .emit_obj(&translation, &types, funcs, tunables, &mut obj)?;
+                    .compile_function(&translation, index, func, tunables, &types)
+            })?
+            .into_iter()
+            .collect();
 
-            // If configured, attempt to use paged memory initialization
-            // instead of the default mode of memory initialization
-            if engine.config().paged_memory_initialization {
-                translation.try_paged_init();
-            }
+        // Collect all the function results into a final ELF object.
+        let mut obj = engine.compiler().object()?;
+        let (funcs, trampolines) =
+            engine
+                .compiler()
+                .emit_obj(&translation, &types, funcs, tunables, &mut obj)?;
 
-            // If configured attempt to use static memory initialization which
-            // can either at runtime be implemented as a single memcpy to
-            // initialize memory or otherwise enabling virtual-memory-tricks
-            // such as mmap'ing from a file to get copy-on-write.
-            if engine.config().memory_init_cow {
-                let align = engine.compiler().page_size_align();
-                let max_always_allowed = engine.config().memory_guaranteed_dense_image_size;
-                translation.try_static_init(align, max_always_allowed);
-            }
+        // If configured, attempt to use paged memory initialization
+        // instead of the default mode of memory initialization
+        if engine.config().paged_memory_initialization {
+            translation.try_paged_init();
+        }
 
-            // Attempt to convert table initializer segments to
-            // FuncTable representation where possible, to enable
-            // table lazy init.
-            translation.try_func_table_init();
+        // If configured attempt to use static memory initialization which
+        // can either at runtime be implemented as a single memcpy to
+        // initialize memory or otherwise enabling virtual-memory-tricks
+        // such as mmap'ing from a file to get copy-on-write.
+        if engine.config().memory_init_cow {
+            let align = engine.compiler().page_size_align();
+            let max_always_allowed = engine.config().memory_guaranteed_dense_image_size;
+            translation.try_static_init(align, max_always_allowed);
+        }
 
-            let (mmap, info) =
-                wasmtime_jit::finish_compile(translation, obj, funcs, trampolines, tunables)?;
-            Ok((mmap, Some(info)))
-        })?;
+        // Attempt to convert table initializer segments to
+        // FuncTable representation where possible, to enable
+        // table lazy init.
+        translation.try_func_table_init();
 
-        Ok((
-            main_module,
-            list,
-            TypeTables {
-                wasm_signatures: types.wasm_signatures,
-                module_signatures: types.module_signatures,
-                instance_signatures: types.instance_signatures,
-            },
-        ))
+        let (mmap, info) =
+            wasmtime_jit::finish_compile(translation, obj, funcs, trampolines, tunables)?;
+
+        Ok((mmap, Some(info), types))
     }
 
     /// Deserializes an in-memory compiled module previously created with
@@ -533,91 +491,35 @@ impl Module {
 
     fn from_parts(
         engine: &Engine,
-        mut modules: Vec<Arc<CompiledModule>>,
-        main_module: usize,
+        mmap: MmapVec,
+        info: Option<CompiledModuleInfo>,
         types: Arc<TypeTables>,
-        module_upvars: &[serialization::SerializedModuleUpvar],
     ) -> Result<Self> {
-        // Validate all modules can be used with the current allocator
-        for module in modules.iter() {
-            engine.allocator().validate(module.module())?;
-        }
+        let module = CompiledModule::from_artifacts(
+            mmap,
+            info,
+            &*engine.config().profiler,
+            engine.unique_id_allocator(),
+        )?;
+
+        // Validate the module can be used with the current allocator
+        engine.allocator().validate(module.module())?;
 
         let signatures = Arc::new(SignatureCollection::new_for_module(
             engine.signatures(),
-            &types.wasm_signatures,
-            modules
-                .iter()
-                .flat_map(|m| m.trampolines().map(|(idx, f, _)| (idx, f))),
+            &types,
+            module.trampolines().map(|(idx, f, _)| (idx, f)),
         ));
 
-        let module = modules.remove(main_module);
-
-        let module_upvars = module_upvars
-            .iter()
-            .map(|m| {
-                mk(
-                    engine,
-                    &modules,
-                    &types,
-                    m.index,
-                    &m.artifact_upvars,
-                    &m.module_upvars,
-                    &signatures,
-                )
-            })
-            .collect();
-
-        return Ok(Self {
+        Ok(Self {
             inner: Arc::new(ModuleInner {
                 engine: engine.clone(),
                 types,
-                module,
-                artifact_upvars: modules,
-                module_upvars,
                 signatures,
                 memory_images: OnceCell::new(),
+                module,
             }),
-        });
-
-        fn mk(
-            engine: &Engine,
-            artifacts: &[Arc<CompiledModule>],
-            types: &Arc<TypeTables>,
-            module_index: usize,
-            artifact_upvars: &[usize],
-            module_upvars: &[serialization::SerializedModuleUpvar],
-            signatures: &Arc<SignatureCollection>,
-        ) -> Module {
-            let module = artifacts[module_index].clone();
-            Module {
-                inner: Arc::new(ModuleInner {
-                    engine: engine.clone(),
-                    types: types.clone(),
-                    module,
-                    memory_images: OnceCell::new(),
-                    artifact_upvars: artifact_upvars
-                        .iter()
-                        .map(|i| artifacts[*i].clone())
-                        .collect(),
-                    module_upvars: module_upvars
-                        .into_iter()
-                        .map(|m| {
-                            mk(
-                                engine,
-                                artifacts,
-                                types,
-                                m.index,
-                                &m.artifact_upvars,
-                                &m.module_upvars,
-                                signatures,
-                            )
-                        })
-                        .collect(),
-                    signatures: signatures.clone(),
-                }),
-            }
-        }
+        })
     }
 
     /// Validates `binary` input data as a WebAssembly binary given the
@@ -654,23 +556,6 @@ impl Module {
         Ok(())
     }
 
-    /// Returns the type signature of this module.
-    pub fn ty(&self) -> ModuleType {
-        let mut sig = ModuleType::new();
-        let env_module = self.compiled_module().module();
-        let types = self.types();
-        for (module, field, ty) in env_module.imports() {
-            sig.add_named_import(module, field, ExternType::from_wasmtime(types, &ty));
-        }
-        for (name, index) in env_module.exports.iter() {
-            sig.add_named_export(
-                name,
-                ExternType::from_wasmtime(types, &env_module.type_of(*index)),
-            );
-        }
-        sig
-    }
-
     /// Serializes this module to a vector of bytes.
     ///
     /// This function is similar to the [`Engine::precompile_module`] method
@@ -686,60 +571,6 @@ impl Module {
         SerializedModule::new(self).to_bytes(&self.inner.engine.config().module_version)
     }
 
-    /// Creates a submodule `Module` value from the specified parameters.
-    ///
-    /// This is used for creating submodules as part of module instantiation.
-    ///
-    /// * `artifact_index` - the index in `artifact_upvars` that we're creating
-    ///   a module for
-    /// * `artifact_upvars` - the mapping of indices of what artifact upvars are
-    ///   needed for the submodule. The length of this array is the length of
-    ///   the upvars array in the submodule to be created, and each element of
-    ///   this array is an index into this module's upvar array.
-    /// * `module_upvars` - similar to `artifact_upvars` this is a mapping of
-    ///   how to create the `module_upvars` of the submodule being created.
-    ///   Each entry in this array is either an index into this module's own
-    ///   module upvars array or it's an index into `modules`, the list of
-    ///   modules so far for the instance where this submodule is being
-    ///   created.
-    /// * `modules` - array indexed by `module_upvars`.
-    ///
-    /// Note that the real meat of this happens in `ModuleEnvironment`
-    /// translation inside of `wasmtime_environ`. This just does the easy thing
-    /// of handling all the indices, over there is where the indices are
-    /// actually calculated and such.
-    pub(crate) fn create_submodule(
-        &self,
-        artifact_index: usize,
-        artifact_upvars: &[usize],
-        module_upvars: &[wasmtime_environ::ModuleUpvar],
-        modules: &PrimaryMap<ModuleIndex, Module>,
-    ) -> Result<Module> {
-        let module = self.inner.artifact_upvars[artifact_index].clone();
-        Ok(Module {
-            inner: Arc::new(ModuleInner {
-                types: self.inner.types.clone(),
-                engine: self.inner.engine.clone(),
-                module,
-                memory_images: OnceCell::new(),
-                artifact_upvars: artifact_upvars
-                    .iter()
-                    .map(|i| self.inner.artifact_upvars[*i].clone())
-                    .collect(),
-                module_upvars: module_upvars
-                    .iter()
-                    .map(|i| match *i {
-                        wasmtime_environ::ModuleUpvar::Inherit(i) => {
-                            self.inner.module_upvars[i].clone()
-                        }
-                        wasmtime_environ::ModuleUpvar::Local(i) => modules[i].clone(),
-                    })
-                    .collect(),
-                signatures: self.inner.signatures.clone(),
-            }),
-        })
-    }
-
     pub(crate) fn compiled_module(&self) -> &Arc<CompiledModule> {
         &self.inner.module
     }
@@ -756,14 +587,6 @@ impl Module {
         &self.inner.signatures
     }
 
-    /// Looks up the module upvar value at the `index` specified.
-    ///
-    /// Note that this panics if `index` is out of bounds since this should
-    /// only be called for valid indices as part of instantiation.
-    pub(crate) fn module_upvar(&self, index: usize) -> &Module {
-        &self.inner.module_upvars[index]
-    }
-
     /// Returns identifier/name that this [`Module`] has. This name
     /// is used in traps/backtrace details.
     ///
@@ -833,7 +656,7 @@ impl Module {
     /// assert_eq!(module.imports().len(), 1);
     /// let import = module.imports().next().unwrap();
     /// assert_eq!(import.module(), "host");
-    /// assert_eq!(import.name(), Some("foo"));
+    /// assert_eq!(import.name(), "foo");
     /// match import.ty() {
     ///     ExternType::Func(_) => { /* ... */ }
     ///     _ => panic!("unexpected import type!"),
@@ -981,6 +804,24 @@ impl Module {
         // statically cast the &Arc<ModuleInner> to &Arc<dyn Trait...>.
         self.inner.clone()
     }
+
+    /// Returns the range of bytes in memory where this module's compilation
+    /// image resides.
+    ///
+    /// The compilation image for a module contains executable code, data, debug
+    /// information, etc. This is roughly the same as the `Module::serialize`
+    /// but not the exact same.
+    ///
+    /// The range of memory reported here is exposed to allow low-level
+    /// manipulation of the memory in platform-specific manners such as using
+    /// `mlock` to force the contents to be paged in immediately or keep them
+    /// paged in after they're loaded.
+    ///
+    /// It is not safe to modify the memory in this range, nor is it safe to
+    /// modify the protections of memory in this range.
+    pub fn image_range(&self) -> Range<usize> {
+        self.compiled_module().image_range()
+    }
 }
 
 fn _assert_send_sync() {
@@ -1034,17 +875,9 @@ impl wasmtime_runtime::ModuleRuntimeInfo for ModuleInner {
     }
 
     fn memory_image(&self, memory: DefinedMemoryIndex) -> Result<Option<&Arc<MemoryImage>>> {
-        if !self.engine.config().memory_init_cow {
-            return Ok(None);
-        }
-
-        let images = self.memory_images.get_or_try_init(|| {
-            ModuleMemoryImages::new(
-                self.module.module(),
-                self.module.wasm_data(),
-                Some(self.module.mmap()),
-            )
-        })?;
+        let images = self
+            .memory_images
+            .get_or_try_init(|| memory_images(&self.engine, &self.module))?;
         Ok(images
             .as_ref()
             .and_then(|images| images.get_memory_image(memory)))
@@ -1158,3 +991,22 @@ impl wasmtime_runtime::ModuleRuntimeInfo for BareModuleInfo {
         }
     }
 }
+
+/// Helper method to construct a `ModuleMemoryImages` for an associated
+/// `CompiledModule`.
+fn memory_images(engine: &Engine, module: &CompiledModule) -> Result<Option<ModuleMemoryImages>> {
+    // If initialization via copy-on-write is explicitly disabled in
+    // configuration then this path is skipped entirely.
+    if !engine.config().memory_init_cow {
+        return Ok(None);
+    }
+
+    // ... otherwise logic is delegated to the `ModuleMemoryImages::new`
+    // constructor.
+    let mmap = if engine.config().force_memory_init_memfd {
+        None
+    } else {
+        Some(module.mmap())
+    };
+    ModuleMemoryImages::new(module.module(), module.wasm_data(), mmap)
+}
