@@ -92,7 +92,7 @@ impl StoreData {
         if id.store_id() != self.id {
             return false;
         }
-        // this should be true as an invariant of our API, but double-check with
+        // This should be true as an invariant of our API, but double-check with
         // debug assertions enabled.
         if cfg!(debug_assertions) {
             assert!(id.index() < T::list(self).len());
@@ -103,6 +103,10 @@ impl StoreData {
     pub(crate) fn funcs(&self) -> impl Iterator<Item = &crate::func::FuncData> {
         self.funcs.iter()
     }
+
+    pub(crate) fn reserve_funcs(&mut self, count: usize) {
+        self.funcs.reserve(count);
+    }
 }
 
 impl<T> Index<Stored<T>> for StoreData
